import logging
from abc import ABC, abstractmethod
from ctypes import *
from functools import partial, wraps
from typing import Protocol, runtime_checkable

import yaml

from pyjrk.pyjrk_properties import PyJrkSettingsProperties
from pyjrk.pyjrk_protocol import jrk_constant as jc
from pyjrk.pyjrk_structures import *


# [J]rk [E]rror [D]ecoder
def JED(func):
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        _e_p = func(*args, **kwargs)
        if bool(_e_p):
            _e = cast(_e_p, POINTER(jrk_error))
            # TODO pass logger to here
            _logger = logging.getLogger("PyJrk")
            _logger.error(_e.contents.message)
            return 1
        else:
            return 0

    return func_wrapper


@runtime_checkable
class LoggerProtocol(Protocol):
    def info(self, message: str, *args, **kwargs) -> None: ...
    def debug(self, message: str, *args, **kwargs) -> None: ...
    def warning(self, message: str, *args, **kwargs) -> None: ...
    def error(self, message: str, *args, **kwargs) -> None: ...


class PyJrkSettingsBase(ABC, PyJrkSettingsProperties):
    """Base class for PyJrk_Settings with static property definitions for IDE support."""

    def __init__(self, device_handle, driver_handles, logger: LoggerProtocol):
        self._device_handle = device_handle
        self.usblib, self.jrklib = driver_handles
        self._logger = logger

        # local vs device - local settings on pc, device settings on jrk
        self._local_settings = jrk_settings()
        self._device_settings = jrk_settings()
        self._device_settings_p = POINTER(jrk_settings)()

        self._convert_structure_to_properties()
        self.auto_apply = False

        self._initialize_settings()

    @abstractmethod
    def _initialize_settings(self): ...

    @abstractmethod
    def _get_jrk_setting_from_device(self, field_name: str, _): ...

    def _set_jrk_setting_with_option(self, field_name, _, value):
        setattr(self._local_settings, field_name, value)
        if self.auto_apply:
            self.apply()

    def _convert_structure_to_properties(self):
        for field_name, field_type in jrk_settings._fields_:
            prop = property(
                fget=partial(self._get_jrk_setting_from_device, field_name),
                fset=partial(self._set_jrk_setting_with_option, field_name),
            )
            setattr(self.__class__, field_name, prop)

    @abstractmethod
    def apply(self): ...

    @abstractmethod
    def print(self): ...

    def load_config(self, config_file):
        with open(config_file, "r") as ymlfile:
            cfg = yaml.safe_load(ymlfile)

        cfg_settings = cfg["jrk_settings"]

        jrk_settings_list = [
            setting_name for setting_name, setting_type in jrk_settings._fields_
        ]

        for setting in cfg_settings:
            if setting in jrk_settings_list:
                if "JRK" in str(cfg_settings[setting]):
                    value = jc[cfg_settings[setting]]
                else:
                    value = cfg_settings[setting]
                setattr(self._local_settings, setting, value)

        if self.auto_apply:
            self.apply()

    ## Wrapped methods from C API
    @JED
    def _get_eeprom_settings(self):
        """Gets the current settings stored in the device's EEPROM memory and write them
        to _device_settings_p.

        This method reads the current settings from the device's EEPROM and stores
        them in _device_settings. This function is always called before calling a
        getting a setting from the device via properties in order to refresh the settings.
        """
        e_p = self.jrklib.jrk_get_eeprom_settings(
            byref(self._device_handle), byref(self._device_settings_p)
        )
        self._device_settings = self._device_settings_p[0]
        return e_p

    @JED
    def _set_eeprom_settings(self):
        """Sets the controller settings based on _local_settings.

        This method writes the previously configured settings stored in _local_settings
        to the device's EEPROM memory. The _local_settings variable must be properly
        set before calling this method.
        """
        e_p = self.jrklib.jrk_set_eeprom_settings(
            byref(self._device_handle), byref(self._local_settings)
        )
        return e_p

    @JED
    def _get_ram_settings(self):
        """Gets the current settings stored in the device's RAM memory and write them
        to _device_settings_p.

        This method reads the current settings from the device's RAM and stores
        them in _device_settings.
        """
        e_p = self.jrklib.jrk_get_ram_settings(
            byref(self._device_handle), byref(self._device_settings_p)
        )
        self._device_settings = self._device_settings_p[0]
        return e_p

    @JED
    def _set_ram_settings(self):
        """Sets the controller settings based on _local_settings.

        This method writes the previously configured settings stored in _local_settings
        to the device's RAM memory. The _local_settings variable must be properly
        set before calling this method.
        """
        e_p = self.jrklib.jrk_set_ram_settings(
            byref(self._device_handle), byref(self._local_settings)
        )
        return e_p

    @JED
    def _jrk_restore_defaults(self):
        """ "Restore factory default settings to EEPROM"""
        e_p = self.jrklib.jrk_restore_defaults(byref(self._device_handle))
        return e_p

    @JED
    def _settings_fix(self):
        warnings_p = POINTER(c_char_p)()
        e_p = self.jrklib.jrk_settings_fix(byref(self._local_settings), warnings_p)
        if bool(warnings_p):
            for w in warnings_p:
                self._logger.warning(w)
        return e_p

    @JED
    def _reinitialize(self):
        e_p = self.jrklib.jrk_reinitialize(byref(self._device_handle))
        return e_p

    @JED
    def _settings_to_string(self, settings_str):
        e_p = self.jrklib.jrk_settings_to_string(
            byref(self._device_settings), byref(settings_str)
        )
        return e_p



from typing import Protocol


class PyJrkSettingsProperties(Protocol):
    """Base class for PyJrk_Settings with static property definitions for IDE support."""

    @property
    def product(self) -> int: ...

    @product.setter
    def product(self, value: int) -> None: ...

    @property
    def firmware_version(self) -> int: ...

    @firmware_version.setter
    def firmware_version(self, value: int) -> None: ...

    @property
    def input_mode(self) -> int: ...

    @input_mode.setter
    def input_mode(self, value: int) -> None: ...

    @property
    def input_error_minimum(self) -> int: ...

    @input_error_minimum.setter
    def input_error_minimum(self, value: int) -> None: ...

    @property
    def input_error_maximum(self) -> int: ...

    @input_error_maximum.setter
    def input_error_maximum(self, value: int) -> None: ...

    @property
    def input_minimum(self) -> int: ...

    @input_minimum.setter
    def input_minimum(self, value: int) -> None: ...

    @property
    def input_maximum(self) -> int: ...

    @input_maximum.setter
    def input_maximum(self, value: int) -> None: ...

    @property
    def input_neutral_minimum(self) -> int: ...

    @input_neutral_minimum.setter
    def input_neutral_minimum(self, value: int) -> None: ...

    @property
    def input_neutral_maximum(self) -> int: ...

    @input_neutral_maximum.setter
    def input_neutral_maximum(self, value: int) -> None: ...

    @property
    def output_minimum(self) -> int: ...

    @output_minimum.setter
    def output_minimum(self, value: int) -> None: ...

    @property
    def output_neutral(self) -> int: ...

    @output_neutral.setter
    def output_neutral(self, value: int) -> None: ...

    @property
    def output_maximum(self) -> int: ...

    @output_maximum.setter
    def output_maximum(self, value: int) -> None: ...

    @property
    def input_invert(self) -> bool: ...

    @input_invert.setter
    def input_invert(self, value: bool) -> None: ...

    @property
    def input_scaling_degree(self) -> int: ...

    @input_scaling_degree.setter
    def input_scaling_degree(self, value: int) -> None: ...

    @property
    def input_detect_disconnect(self) -> bool: ...

    @input_detect_disconnect.setter
    def input_detect_disconnect(self, value: bool) -> None: ...

    @property
    def input_analog_samples_exponent(self) -> int: ...

    @input_analog_samples_exponent.setter
    def input_analog_samples_exponent(self, value: int) -> None: ...

    @property
    def feedback_mode(self) -> int: ...

    @feedback_mode.setter
    def feedback_mode(self, value: int) -> None: ...

    @property
    def feedback_error_minimum(self) -> int: ...

    @feedback_error_minimum.setter
    def feedback_error_minimum(self, value: int) -> None: ...

    @property
    def feedback_error_maximum(self) -> int: ...

    @feedback_error_maximum.setter
    def feedback_error_maximum(self, value: int) -> None: ...

    @property
    def feedback_minimum(self) -> int: ...

    @feedback_minimum.setter
    def feedback_minimum(self, value: int) -> None: ...

    @property
    def feedback_maximum(self) -> int: ...

    @feedback_maximum.setter
    def feedback_maximum(self, value: int) -> None: ...

    @property
    def feedback_invert(self) -> bool: ...

    @feedback_invert.setter
    def feedback_invert(self, value: bool) -> None: ...

    @property
    def feedback_detect_disconnect(self) -> bool: ...

    @feedback_detect_disconnect.setter
    def feedback_detect_disconnect(self, value: bool) -> None: ...

    @property
    def feedback_dead_zone(self) -> int: ...

    @feedback_dead_zone.setter
    def feedback_dead_zone(self, value: int) -> None: ...

    @property
    def feedback_analog_samples_exponent(self) -> int: ...

    @feedback_analog_samples_exponent.setter
    def feedback_analog_samples_exponent(self, value: int) -> None: ...

    @property
    def feedback_wraparound(self) -> bool: ...

    @feedback_wraparound.setter
    def feedback_wraparound(self, value: bool) -> None: ...

    @property
    def serial_mode(self) -> int: ...

    @serial_mode.setter
    def serial_mode(self, value: int) -> None: ...

    @property
    def serial_baud_rate(self) -> int: ...

    @serial_baud_rate.setter
    def serial_baud_rate(self, value: int) -> None: ...

    @property
    def serial_timeout(self) -> int: ...

    @serial_timeout.setter
    def serial_timeout(self, value: int) -> None: ...

    @property
    def serial_device_number(self) -> int: ...

    @serial_device_number.setter
    def serial_device_number(self, value: int) -> None: ...

    @property
    def never_sleep(self) -> bool: ...

    @never_sleep.setter
    def never_sleep(self, value: bool) -> None: ...

    @property
    def serial_enable_crc(self) -> bool: ...

    @serial_enable_crc.setter
    def serial_enable_crc(self, value: bool) -> None: ...

    @property
    def serial_enable_14bit_device_number(self) -> bool: ...

    @serial_enable_14bit_device_number.setter
    def serial_enable_14bit_device_number(self, value: bool) -> None: ...

    @property
    def serial_disable_compact_protocol(self) -> bool: ...

    @serial_disable_compact_protocol.setter
    def serial_disable_compact_protocol(self, value: bool) -> None: ...

    @property
    def proportional_multiplier(self) -> int: ...

    @proportional_multiplier.setter
    def proportional_multiplier(self, value: int) -> None: ...

    @property
    def proportional_exponent(self) -> int: ...

    @proportional_exponent.setter
    def proportional_exponent(self, value: int) -> None: ...

    @property
    def integral_multiplier(self) -> int: ...

    @integral_multiplier.setter
    def integral_multiplier(self, value: int) -> None: ...

    @property
    def integral_exponent(self) -> int: ...

    @integral_exponent.setter
    def integral_exponent(self, value: int) -> None: ...

    @property
    def derivative_multiplier(self) -> int: ...

    @derivative_multiplier.setter
    def derivative_multiplier(self, value: int) -> None: ...

    @property
    def derivative_exponent(self) -> int: ...

    @derivative_exponent.setter
    def derivative_exponent(self, value: int) -> None: ...

    @property
    def pid_period(self) -> int: ...

    @pid_period.setter
    def pid_period(self, value: int) -> None: ...

    @property
    def integral_divider_exponent(self) -> int: ...

    @integral_divider_exponent.setter
    def integral_divider_exponent(self, value: int) -> None: ...

    @property
    def integral_limit(self) -> int: ...

    @integral_limit.setter
    def integral_limit(self, value: int) -> None: ...

    @property
    def reset_integral(self) -> bool: ...

    @reset_integral.setter
    def reset_integral(self, value: bool) -> None: ...

    @property
    def pwm_frequency(self) -> int: ...

    @pwm_frequency.setter
    def pwm_frequency(self, value: int) -> None: ...

    @property
    def current_samples_exponent(self) -> int: ...

    @current_samples_exponent.setter
    def current_samples_exponent(self, value: int) -> None: ...

    @property
    def hard_overcurrent_threshold(self) -> int: ...

    @hard_overcurrent_threshold.setter
    def hard_overcurrent_threshold(self, value: int) -> None: ...

    @property
    def current_offset_calibration(self) -> int: ...

    @current_offset_calibration.setter
    def current_offset_calibration(self, value: int) -> None: ...

    @property
    def current_scale_calibration(self) -> int: ...

    @current_scale_calibration.setter
    def current_scale_calibration(self, value: int) -> None: ...

    @property
    def motor_invert(self) -> bool: ...

    @motor_invert.setter
    def motor_invert(self, value: bool) -> None: ...

    @property
    def max_duty_cycle_while_feedback_out_of_range(self) -> int: ...

    @max_duty_cycle_while_feedback_out_of_range.setter
    def max_duty_cycle_while_feedback_out_of_range(self, value: int) -> None: ...

    @property
    def max_acceleration_forward(self) -> int: ...

    @max_acceleration_forward.setter
    def max_acceleration_forward(self, value: int) -> None: ...

    @property
    def max_acceleration_reverse(self) -> int: ...

    @max_acceleration_reverse.setter
    def max_acceleration_reverse(self, value: int) -> None: ...

    @property
    def max_deceleration_forward(self) -> int: ...

    @max_deceleration_forward.setter
    def max_deceleration_forward(self, value: int) -> None: ...

    @property
    def max_deceleration_reverse(self) -> int: ...

    @max_deceleration_reverse.setter
    def max_deceleration_reverse(self, value: int) -> None: ...

    @property
    def max_duty_cycle_forward(self) -> int: ...

    @max_duty_cycle_forward.setter
    def max_duty_cycle_forward(self, value: int) -> None: ...

    @property
    def max_duty_cycle_reverse(self) -> int: ...

    @max_duty_cycle_reverse.setter
    def max_duty_cycle_reverse(self, value: int) -> None: ...

    @property
    def encoded_hard_current_limit_forward(self) -> int: ...

    @encoded_hard_current_limit_forward.setter
    def encoded_hard_current_limit_forward(self, value: int) -> None: ...

    @property
    def encoded_hard_current_limit_reverse(self) -> int: ...

    @encoded_hard_current_limit_reverse.setter
    def encoded_hard_current_limit_reverse(self, value: int) -> None: ...

    @property
    def brake_duration_forward(self) -> int: ...

    @brake_duration_forward.setter
    def brake_duration_forward(self, value: int) -> None: ...

    @property
    def brake_duration_reverse(self) -> int: ...

    @brake_duration_reverse.setter
    def brake_duration_reverse(self, value: int) -> None: ...

    @property
    def soft_current_limit_forward(self) -> int: ...

    @soft_current_limit_forward.setter
    def soft_current_limit_forward(self, value: int) -> None: ...

    @property
    def soft_current_limit_reverse(self) -> int: ...

    @soft_current_limit_reverse.setter
    def soft_current_limit_reverse(self, value: int) -> None: ...

    @property
    def soft_current_regulation_level_forward(self) -> int: ...

    @soft_current_regulation_level_forward.setter
    def soft_current_regulation_level_forward(self, value: int) -> None: ...

    @property
    def soft_current_regulation_level_reverse(self) -> int: ...

    @soft_current_regulation_level_reverse.setter
    def soft_current_regulation_level_reverse(self, value: int) -> None: ...

    @property
    def coast_when_off(self) -> bool: ...

    @coast_when_off.setter
    def coast_when_off(self, value: bool) -> None: ...

    @property
    def error_enable(self) -> int: ...

    @error_enable.setter
    def error_enable(self, value: int) -> None: ...

    @property
    def error_latch(self) -> int: ...

    @error_latch.setter
    def error_latch(self, value: int) -> None: ...

    @property
    def error_hard(self) -> int: ...

    @error_hard.setter
    def error_hard(self, value: int) -> None: ...

    @property
    def vin_calibration(self) -> int: ...

    @vin_calibration.setter
    def vin_calibration(self, value: int) -> None: ...

    @property
    def disable_i2c_pullups(self) -> bool: ...

    @disable_i2c_pullups.setter
    def disable_i2c_pullups(self, value: bool) -> None: ...

    @property
    def analog_sda_pullup(self) -> bool: ...

    @analog_sda_pullup.setter
    def analog_sda_pullup(self, value: bool) -> None: ...

    @property
    def always_analog_sda(self) -> bool: ...

    @always_analog_sda.setter
    def always_analog_sda(self, value: bool) -> None: ...

    @property
    def always_analog_fba(self) -> bool: ...

    @always_analog_fba.setter
    def always_analog_fba(self, value: bool) -> None: ...

    @property
    def fbt_method(self) -> int: ...

    @fbt_method.setter
    def fbt_method(self, value: int) -> None: ...

    @property
    def fbt_timing_clock(self) -> int: ...

    @fbt_timing_clock.setter
    def fbt_timing_clock(self, value: int) -> None: ...

    @property
    def fbt_timing_polarity(self) -> bool: ...

    @fbt_timing_polarity.setter
    def fbt_timing_polarity(self, value: bool) -> None: ...

    @property
    def fbt_timing_timeout(self) -> int: ...

    @fbt_timing_timeout.setter
    def fbt_timing_timeout(self, value: int) -> None: ...

    @property
    def fbt_samples(self) -> int: ...

    @fbt_samples.setter
    def fbt_samples(self, value: int) -> None: ...

    @property
    def fbt_divider_exponent(self) -> int: ...

    @fbt_divider_exponent.setter
    def fbt_divider_exponent(self, value: int) -> None: ...


class PyJrkVariablesProperties(Protocol):
    """Base class for PyJrkVariables with static property definitions for IDE support."""

    @property
    def input(self) -> int: ...

    @input.setter
    def input(self, value: int) -> None: ...

    @property
    def target(self) -> int: ...

    @target.setter
    def target(self, value: int) -> None: ...

    @property
    def feedback(self) -> int: ...

    @feedback.setter
    def feedback(self, value: int) -> None: ...

    @property
    def scaled_feedback(self) -> int: ...

    @scaled_feedback.setter
    def scaled_feedback(self, value: int) -> None: ...

    @property
    def integral(self) -> int: ...

    @integral.setter
    def integral(self, value: int) -> None: ...

    @property
    def duty_cycle_target(self) -> int: ...

    @duty_cycle_target.setter
    def duty_cycle_target(self, value: int) -> None: ...

    @property
    def duty_cycle(self) -> int: ...

    @duty_cycle.setter
    def duty_cycle(self, value: int) -> None: ...

    @property
    def current_low_res(self) -> int: ...

    @current_low_res.setter
    def current_low_res(self, value: int) -> None: ...

    @property
    def pid_period_exceeded(self) -> bool: ...

    @pid_period_exceeded.setter
    def pid_period_exceeded(self, value: bool) -> None: ...

    @property
    def pid_period_count(self) -> int: ...

    @pid_period_count.setter
    def pid_period_count(self, value: int) -> None: ...

    @property
    def error_flags_halting(self) -> int: ...

    @error_flags_halting.setter
    def error_flags_halting(self, value: int) -> None: ...

    @property
    def error_flags_occurred(self) -> int: ...

    @error_flags_occurred.setter
    def error_flags_occurred(self, value: int) -> None: ...

    @property
    def vin_voltage(self) -> int: ...

    @vin_voltage.setter
    def vin_voltage(self, value: int) -> None: ...

    @property
    def current(self) -> int: ...

    @current.setter
    def current(self, value: int) -> None: ...

    @property
    def device_reset(self) -> int: ...

    @device_reset.setter
    def device_reset(self, value: int) -> None: ...

    @property
    def up_time(self) -> int: ...

    @up_time.setter
    def up_time(self, value: int) -> None: ...

    @property
    def rc_pulse_width(self) -> int: ...

    @rc_pulse_width.setter
    def rc_pulse_width(self, value: int) -> None: ...

    @property
    def fbt_reading(self) -> int: ...

    @fbt_reading.setter
    def fbt_reading(self, value: int) -> None: ...

    @property
    def raw_current(self) -> int: ...

    @raw_current.setter
    def raw_current(self, value: int) -> None: ...

    @property
    def encoded_hard_current_limit(self) -> int: ...

    @encoded_hard_current_limit.setter
    def encoded_hard_current_limit(self, value: int) -> None: ...

    @property
    def last_duty_cycle(self) -> int: ...

    @last_duty_cycle.setter
    def last_duty_cycle(self, value: int) -> None: ...

    @property
    def current_chopping_consecutive_count(self) -> int: ...

    @current_chopping_consecutive_count.setter
    def current_chopping_consecutive_count(self, value: int) -> None: ...

    @property
    def current_chopping_occurrence_count(self) -> int: ...

    @current_chopping_occurrence_count.setter
    def current_chopping_occurrence_count(self, value: int) -> None: ...

    @property
    def force_mode(self) -> int: ...

    @force_mode.setter
    def force_mode(self, value: int) -> None: ...



jrk_constant = {
    # Present in jrk.h
    "JRK_PRODUCT_UMC04A_30V": 1,
    "JRK_PRODUCT_UMC04A_40V": 2,
    "JRK_PRODUCT_UMC05A_30V": 3,
    "JRK_PRODUCT_UMC05A_40V": 4,
    "JRK_PRODUCT_UMC06A":    5,

    # Present in jrk_protocol.h
    'JRK_USB_VENDOR_ID': 0x1FFB,
    'JRK_USB_PRODUCT_ID_UMC04A_30V': 0x00B7,
    'JRK_USB_PRODUCT_ID_UMC04A_40V': 0x00B9,
    'JRK_USB_PRODUCT_ID_UMC05A_30V': 0x00BF,
    'JRK_USB_PRODUCT_ID_UMC05A_40V': 0x00C1,
    'JRK_USB_PRODUCT_ID_UMC06A': 0x00C5,
    
    'JRK_CMD_REINITIALIZE': 0x10,
    'JRK_CMD_SET_EEPROM_SETTING': 0x13,
    'JRK_CMD_GET_DEBUG_DATA': 0x20,
    'JRK_CMD_GET_VARIABLE_SERIAL': 0x80,
    'JRK_CMD_GET_ERROR_FLAGS_HALTING_SERIAL': 0xB3,
    'JRK_CMD_GET_ERROR_FLAGS_OCCURRED_SERIAL': 0xB5,
    'JRK_CMD_SET_TARGET_USB': 0x84,
    'JRK_CMD_STOP_MOTOR_USB': 0x87,
    'JRK_CMD_SET_TARGET_SERIAL': 0xC0,
    'JRK_CMD_SET_TARGET_LOW_RES_REV': 0xE0,
    'JRK_CMD_SET_TARGET_LOW_RES_FWD': 0xE1,
    'JRK_CMD_GET_EEPROM_SETTINGS': 0xE3,
    'JRK_CMD_GET_VARIABLES': 0xE5,
    'JRK_CMD_SET_RAM_SETTINGS': 0xE6,
    'JRK_CMD_GET_RAM_SETTINGS': 0xEA,
    'JRK_CMD_GET_CURRENT_CHOPPING_OCCURRENCE_COUNT': 0xEC,
    'JRK_CMD_FORCE_DUTY_CYCLE_TARGET': 0xF2,
    'JRK_CMD_FORCE_DUTY_CYCLE': 0xF4,
    'JRK_CMD_STOP_MOTOR_SERIAL': 0xFF,
    'JRK_CMD_START_BOOTLOADER': 0xFF,
    
    'JRK_CMD_GET_VARIABLE_SERIAL_MASK': 0xC0,
    'JRK_CMD_SET_TARGET_SERIAL_MASK': 0xE0,
    
    'JRK_GET_VARIABLES_FLAG_CLEAR_ERROR_FLAGS_HALTING': 0,
    'JRK_GET_VARIABLES_FLAG_CLEAR_ERROR_FLAGS_OCCURRED': 1,
    'JRK_GET_VARIABLES_FLAG_CLEAR_CURRENT_CHOPPING_OCCURRENCE_COUNT': 2,
    
    'JRK_REINITIALIZE_FLAG_PRESERVE_ERRORS': 0,
    
    'JRK_VAR_INPUT': 0x00,
    'JRK_VAR_TARGET': 0x02,
    'JRK_VAR_FEEDBACK': 0x04,
    'JRK_VAR_SCALED_FEEDBACK': 0x06,
    'JRK_VAR_INTEGRAL': 0x08,
    'JRK_VAR_DUTY_CYCLE_TARGET': 0x0A,
    'JRK_VAR_DUTY_CYCLE': 0x0C,
    'JRK_VAR_CURRENT_LOW_RES': 0x0E,
    'JRK_VAR_PID_PERIOD_EXCEEDED': 0x0F,
    'JRK_VAR_PID_PERIOD_COUNT': 0x10,
    'JRK_VAR_ERROR_FLAGS_HALTING': 0x12,
    'JRK_VAR_ERROR_FLAGS_OCCURRED': 0x14,
    'JRK_VAR_FLAG_BYTE1': 0x16,
    'JRK_VAR_VIN_VOLTAGE': 0x17,
    'JRK_VAR_CURRENT': 0x19,
    'JRK_VAR_DEVICE_RESET': 0x1F,
    'JRK_VAR_UP_TIME': 0x20,
    'JRK_VAR_RC_PULSE_WIDTH': 0x24,
    'JRK_VAR_FBT_READING': 0x26,
    'JRK_VAR_ANALOG_READING_SDA': 0x28,
    'JRK_VAR_ANALOG_READING_FBA': 0x2A,
    'JRK_VAR_DIGITAL_READINGS': 0x2C,
    'JRK_VAR_RAW_CURRENT': 0x2D,
    'JRK_VAR_ENCODED_HARD_CURRENT_LIMIT': 0x2F,
    'JRK_VAR_LAST_DUTY_CYCLE': 0x31,
    'JRK_VAR_CURRENT_CHOPPING_CONSECUTIVE_COUNT': 0x33,
    'JRK_VAR_CURRENT_CHOPPING_OCCURRENCE_COUNT': 0x34,
    'JRK_VARIABLES_SIZE': 0x35,
    
    'JRK_SETTING_NOT_INITIALIZED': 0x00,
    'JRK_SETTING_OPTIONS_BYTE1': 0x01,
    'JRK_SETTING_OPTIONS_BYTE2': 0x02,
    'JRK_SETTING_INPUT_MODE': 0x03,
    'JRK_SETTING_INPUT_ERROR_MINIMUM': 0x04,
    'JRK_SETTING_INPUT_ERROR_MAXIMUM': 0x06,
    'JRK_SETTING_INPUT_MINIMUM': 0x08,
    'JRK_SETTING_INPUT_MAXIMUM': 0x0A,
    'JRK_SETTING_INPUT_NEUTRAL_MINIMUM': 0x0C,
    'JRK_SETTING_INPUT_NEUTRAL_MAXIMUM': 0x0E,
    'JRK_SETTING_OUTPUT_MINIMUM': 0x10,
    'JRK_SETTING_OUTPUT_NEUTRAL': 0x12,
    'JRK_SETTING_OUTPUT_MAXIMUM': 0x14,
    'JRK_SETTING_INPUT_SCALING_DEGREE': 0x16,
    'JRK_SETTING_INPUT_ANALOG_SAMPLES_EXPONENT': 0x17,
    'JRK_SETTING_FEEDBACK_MODE': 0x18,
    'JRK_SETTING_FEEDBACK_ERROR_MINIMUM': 0x19,
    'JRK_SETTING_FEEDBACK_ERROR_MAXIMUM': 0x1B,
    'JRK_SETTING_FEEDBACK_MINIMUM': 0x1D,
    'JRK_SETTING_FEEDBACK_MAXIMUM': 0x1F,
    'JRK_SETTING_FEEDBACK_DEAD_ZONE': 0x21,
    'JRK_SETTING_FEEDBACK_ANALOG_SAMPLES_EXPONENT': 0x22,
    'JRK_SETTING_SERIAL_MODE': 0x23,
    'JRK_SETTING_SERIAL_BAUD_RATE_GENERATOR': 0x24,
    'JRK_SETTING_SERIAL_TIMEOUT': 0x26,
    'JRK_SETTING_SERIAL_DEVICE_NUMBER': 0x28,
    'JRK_SETTING_ERROR_ENABLE': 0x2A,
    'JRK_SETTING_ERROR_LATCH': 0x2C,
    'JRK_SETTING_ERROR_HARD': 0x2E,
    'JRK_SETTING_VIN_CALIBRATION': 0x30,
    'JRK_SETTING_PWM_FREQUENCY': 0x32,
    'JRK_SETTING_CURRENT_SAMPLES_EXPONENT': 0x33,
    'JRK_SETTING_HARD_OVERCURRENT_THRESHOLD': 0x34,
    'JRK_SETTING_CURRENT_OFFSET_CALIBRATION': 0x35,
    'JRK_SETTING_CURRENT_SCALE_CALIBRATION': 0x37,
    'JRK_SETTING_FBT_METHOD': 0x39,
    'JRK_SETTING_FBT_OPTIONS': 0x3A,
    'JRK_SETTING_FBT_TIMING_TIMEOUT': 0x3B,
    'JRK_SETTING_FBT_SAMPLES': 0x3D,
    'JRK_SETTING_FBT_DIVIDER_EXPONENT': 0x3E,
    'JRK_SETTING_INTEGRAL_DIVIDER_EXPONENT': 0x3F,
    'JRK_SETTING_SOFT_CURRENT_REGULATION_LEVEL_FORWARD': 0x40,
    'JRK_SETTING_SOFT_CURRENT_REGULATION_LEVEL_REVERSE': 0x42,
    'JRK_SETTING_OPTIONS_BYTE3': 0x50,
    'JRK_SETTING_PROPORTIONAL_MULTIPLIER': 0x51,
    'JRK_SETTING_PROPORTIONAL_EXPONENT': 0x53,
    'JRK_SETTING_INTEGRAL_MULTIPLIER': 0x54,
    'JRK_SETTING_INTEGRAL_EXPONENT': 0x56,
    'JRK_SETTING_DERIVATIVE_MULTIPLIER': 0x57,
    'JRK_SETTING_DERIVATIVE_EXPONENT': 0x59,
    'JRK_SETTING_PID_PERIOD': 0x5A,
    'JRK_SETTING_INTEGRAL_LIMIT': 0x5C,
    'JRK_SETTING_MAX_DUTY_CYCLE_WHILE_FEEDBACK_OUT_OF_RANGE': 0x5E,
    'JRK_SETTING_MAX_ACCELERATION_FORWARD': 0x60,
    'JRK_SETTING_MAX_ACCELERATION_REVERSE': 0x62,
    'JRK_SETTING_MAX_DECELERATION_FORWARD': 0x64,
    'JRK_SETTING_MAX_DECELERATION_REVERSE': 0x66,
    'JRK_SETTING_MAX_DUTY_CYCLE_FORWARD': 0x68,
    'JRK_SETTING_MAX_DUTY_CYCLE_REVERSE': 0x6A,
    'JRK_SETTING_ENCODED_HARD_CURRENT_LIMIT_FORWARD': 0x6C,
    'JRK_SETTING_ENCODED_HARD_CURRENT_LIMIT_REVERSE': 0x6E,
    'JRK_SETTING_BRAKE_DURATION_FORWARD': 0x70,
    'JRK_SETTING_BRAKE_DURATION_REVERSE': 0x71,
    'JRK_SETTING_SOFT_CURRENT_LIMIT_FORWARD': 0x72,
    'JRK_SETTING_SOFT_CURRENT_LIMIT_REVERSE': 0x74,
    
    'JRK_SETTINGS_SIZE': 0x76,
    
    'JRK_OPTIONS_BYTE1_NEVER_SLEEP': 0,
    'JRK_OPTIONS_BYTE1_SERIAL_ENABLE_CRC': 1,
    'JRK_OPTIONS_BYTE1_SERIAL_ENABLE_14BIT_DEVICE_NUMBER': 2,
    'JRK_OPTIONS_BYTE1_SERIAL_DISABLE_COMPACT_PROTOCOL': 3,
    'JRK_OPTIONS_BYTE1_DISABLE_I2C_PULLUPS': 4,
    'JRK_OPTIONS_BYTE1_ANALOG_SDA_PULLUP': 5,
    'JRK_OPTIONS_BYTE1_ALWAYS_ANALOG_SDA': 6,
    'JRK_OPTIONS_BYTE1_ALWAYS_ANALOG_FBA': 7,
    
    'JRK_OPTIONS_BYTE2_INPUT_INVERT': 0,
    'JRK_OPTIONS_BYTE2_INPUT_DETECT_DISCONNECT': 1,
    'JRK_OPTIONS_BYTE2_FEEDBACK_INVERT': 2,
    'JRK_OPTIONS_BYTE2_FEEDBACK_DETECT_DISCONNECT': 3,
    'JRK_OPTIONS_BYTE2_FEEDBACK_WRAPAROUND': 4,
    'JRK_OPTIONS_BYTE2_MOTOR_INVERT': 5,
    
    'JRK_OPTIONS_BYTE3_RESET_INTEGRAL': 0,
    'JRK_OPTIONS_BYTE3_COAST_WHEN_OFF': 1,
    
    'JRK_FBT_OPTIONS_TIMING_POLARITY': 0,
    'JRK_FBT_OPTIONS_TIMING_CLOCK': 4,
    'JRK_FBT_OPTIONS_TIMING_CLOCK_MASK': 0x7,
    
    'JRK_FIRMWARE_MODIFICATION_STRING_INDEX': 6,
    
    'JRK_MAX_USB_RESPONSE_SIZE': 128,
    
    'JRK_MAX_ALLOWED_DUTY_CYCLE': 600,
    
    'JRK_INPUT_MODE_SERIAL': 0,
    'JRK_INPUT_MODE_ANALOG': 1,
    'JRK_INPUT_MODE_RC': 2,
    
    'JRK_FEEDBACK_MODE_NONE': 0,
    'JRK_FEEDBACK_MODE_ANALOG': 1,
    'JRK_FEEDBACK_MODE_FREQUENCY': 2,
    
    'JRK_SERIAL_MODE_USB_DUAL_PORT': 0,
    'JRK_SERIAL_MODE_USB_CHAINED': 1,
    'JRK_SERIAL_MODE_UART': 2,
    
    'JRK_PWM_FREQUENCY_20': 0,
    'JRK_PWM_FREQUENCY_5': 1,
    
    'JRK_SCALING_DEGREE_LINEAR': 0,
    'JRK_SCALING_DEGREE_QUADRATIC': 1,
    'JRK_SCALING_DEGREE_CUBIC': 2,
    'JRK_SCALING_DEGREE_QUARTIC': 3,
    'JRK_SCALING_DEGREE_QUINTIC': 4,
    
    'JRK_BRAKE_DURATION_UNITS': 5,
    'JRK_MAX_ALLOWED_BRAKE_DURATION': 1275,
    
    'JRK_SERIAL_TIMEOUT_UNITS': 10,
    'JRK_MAX_ALLOWED_SERIAL_TIMEOUT': 655350,
    
    'JRK_BAUD_RATE_GENERATOR_FACTOR': 12000000,
    
    'JRK_MIN_ALLOWED_BAUD_RATE': 184,
    'JRK_MAX_ALLOWED_BAUD_RATE': 115385,
    
    'JRK_FBT_METHOD_PULSE_COUNTING': 0,
    'JRK_FBT_METHOD_PULSE_TIMING': 1,
    
    'JRK_MAX_ALLOWED_FBT_SAMPLES': 32,
    
    'JRK_FBT_TIMING_CLOCK_12': 0,
    'JRK_FBT_TIMING_CLOCK_6': 1,
    'JRK_FBT_TIMING_CLOCK_3': 2,
    'JRK_FBT_TIMING_CLOCK_1_5': 3,
    'JRK_FBT_TIMING_CLOCK_48': 4,
    'JRK_FBT_TIMING_CLOCK_24': 5,
    
    'JRK_FORCE_MODE_NONE': 0,
    'JRK_FORCE_MODE_DUTY_CYCLE_TARGET': 1,
    'JRK_FORCE_MODE_DUTY_CYCLE': 2,
    
    'JRK_RESET_POWER_UP': 0,
    'JRK_RESET_BROWNOUT': 1,
    'JRK_RESET_RESET_LINE': 2,
    'JRK_RESET_WATCHDOG': 4,
    'JRK_RESET_SOFTWARE': 8,
    'JRK_RESET_STACK_OVERFLOW': 16,
    'JRK_RESET_STACK_UNDERFLOW': 32,
    
    'JRK_ERROR_AWAITING_COMMAND': 0,
    'JRK_ERROR_NO_POWER': 1,
    'JRK_ERROR_MOTOR_DRIVER': 2,
    'JRK_ERROR_INPUT_INVALID': 3,
    'JRK_ERROR_INPUT_DISCONNECT': 4,
    'JRK_ERROR_FEEDBACK_DISCONNECT': 5,
    'JRK_ERROR_SOFT_OVERCURRENT': 6,
    'JRK_ERROR_SERIAL_SIGNAL': 7,
    'JRK_ERROR_SERIAL_OVERRUN': 8,
    'JRK_ERROR_SERIAL_BUFFER_FULL': 9,
    'JRK_ERROR_SERIAL_CRC': 10,
    'JRK_ERROR_SERIAL_PROTOCOL': 11,
    'JRK_ERROR_SERIAL_TIMEOUT': 12,
    'JRK_ERROR_HARD_OVERCURRENT': 13,
    
    'JRK_PIN_NUM_SCL': 0,
    'JRK_PIN_NUM_SDA': 1,
    'JRK_PIN_NUM_TX': 2,
    'JRK_PIN_NUM_RX': 3,
    'JRK_PIN_NUM_RC': 4,
    'JRK_PIN_NUM_AUX': 5,
    'JRK_PIN_NUM_FBA': 6,
    'JRK_PIN_NUM_FBT': 7,
    'JRK_CONTROL_PIN_COUNT': 8,
}


import logging
import os
import platform
from ctypes import *
from functools import partial
from typing import Callable

from pyjrk.pyjrk_base import JED, LoggerProtocol, PyJrkSettingsBase
from pyjrk.pyjrk_properties import PyJrkVariablesProperties
from pyjrk.pyjrk_protocol import jrk_constant as jc
from pyjrk.pyjrk_structures import *


class PyJrk:
    # Type annotations for dynamically created methods
    set_target: Callable[[int], int]
    stop_motor: Callable[[], int]
    force_duty_cycle_target: Callable[[], int]
    force_duty_cycle: Callable[[], int]
    reinitialize: Callable[[int], int]

    def __init__(self, logger: LoggerProtocol = None):
        self._logger = logger if logger else self._initialize_default_logger()
        self._load_drivers()

        self.device = None
        self.handle = None
        self.eeprom_settings: PyJrkEEPROMSettings = None
        self.ram_settings: PyJrkRAMSettings = None
        self.variables: PyJrkVariables = None
        self._commands = [
            ("set_target", c_uint16),
            ("stop_motor", None),
            ("force_duty_cycle_target", None),
            ("force_duty_cycle", None),
            ("reinitialize", c_uint8),
        ]
        self._create_jrk_command_attributes()

    def _initialize_default_logger(self):
        # - Logging -
        self._log_level = logging.DEBUG
        _logger = logging.getLogger("PyJrk")
        _logger.setLevel(self._log_level)
        _formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        # Console Logging
        _ch = logging.StreamHandler()
        _ch.setLevel(self._log_level)
        _ch.setFormatter(_formatter)
        _logger.addHandler(_ch)
        return _logger

    @property
    def log_level(self):
        return self._log_level

    @log_level.setter
    def log_level(self, level):
        self._log_level = level
        self._logger.setLevel(level)

    def _load_drivers(self):
        # Driver Locations (x64)
        file_path = os.path.dirname(os.path.abspath(__file__))
        if platform.system() == "Windows":
            # Windows DLL paths
            self.usblib = windll.LoadLibrary(
                file_path + "\\drivers\\x64\\libusbp-1.dll"
            )  # type: ignore
            self.jrklib = windll.LoadLibrary(
                file_path + "\\drivers\\x64\\libpololu-jrk2-1.dll"
            )  # type: ignore
        elif platform.system() == "Linux":
            # Linux shared library paths
            self.usblib = CDLL(file_path + "/drivers/linux/libusbp-1.so")
            self.jrklib = CDLL(file_path + "/drivers/linux/libpololu-jrk2-1.so")
        self._logger.debug("JRK Drivers loaded")

    def _create_jrk_command_attributes(self):
        for cmd_name, value_c_type in self._commands:
            if bool(value_c_type):
                setattr(
                    self.__class__,
                    cmd_name,
                    partial(self._jrk_command_with_value, cmd_name, value_c_type),
                )
            else:
                setattr(self.__class__, cmd_name, partial(self._jrk_command, cmd_name))

    @JED
    def _jrk_command(self, cmd_name):
        e_p = getattr(self.jrklib, "jrk_" + cmd_name)(byref(self.handle))
        return e_p

    @JED
    def _jrk_command_with_value(self, cmd_name, value_c_type, value):
        if "JRK" in str(value):
            value = jc[value]
        e_p = getattr(self.jrklib, "jrk_" + cmd_name)(
            byref(self.handle), value_c_type(value)
        )
        return e_p

    @JED
    def _list_connected_devices(self):
        self._devcnt = c_size_t(0)
        self._dev_pp = POINTER(POINTER(jrk_device))()
        e_p = self.jrklib.jrk_list_connected_devices(
            byref(self._dev_pp), byref(self._devcnt)
        )
        return e_p

    @JED
    def _jrk_handle_open(self):
        handle_p = POINTER(jrk_handle)()
        e_p = self.jrklib.jrk_handle_open(byref(self.device), byref(handle_p))
        self.handle = handle_p[0]
        return e_p

    def list_connected_device_serial_numbers(self):
        self._list_connected_devices()
        jrk_list = []
        if not self._devcnt.value:
            self._logger.warning("No Jrk devices connected.")
        for i in range(0, self._devcnt.value):
            jrkdev = self._dev_pp[0][i]
            jrk_list.append(jrkdev.serial_number.decode("utf-8"))
        return jrk_list

    def connect_to_serial_number(self, serial_number):
        self._list_connected_devices()
        for i in range(0, self._devcnt.value):
            if serial_number == self._dev_pp[0][i].serial_number.decode("utf-8"):
                self.device = self._dev_pp[0][i]
                self._jrk_handle_open()
                self.variables = PyJrkVariables(
                    self.handle, (self.usblib, self.jrklib), self._logger
                )
                self.eeprom_settings = PyJrkEEPROMSettings(
                    self.handle, (self.usblib, self.jrklib), self._logger
                )
                self.ram_settings = PyJrkRAMSettings(
                    self.handle, (self.usblib, self.jrklib), self._logger
                )
                return 0
        if not self.device:
            self._logger.error("Serial number device not found.")
            return 1


class PyJrkVariables(PyJrkVariablesProperties):
    def __init__(self, device_handle, driver_handles, logger: LoggerProtocol):
        self._device_handle = device_handle
        self.usblib, self.jrklib = driver_handles
        self._logger = logger

        self._jrk_variables_p = POINTER(jrk_variables)()
        self._jrk_variables = jrk_variables()

        self.pin_info = []
        for i in range(0, jc["JRK_CONTROL_PIN_COUNT"]):
            self.pin_info.append(type("pinfo_" + str(i), (object,), {})())

        self._convert_structure_to_readonly_properties()

    def _convert_structure_to_readonly_properties(self):
        for field_name, field_type in jrk_variables._fields_:
            if not field_name == "pin_info":
                prop = property(
                    fget=partial(self._get_jrk_readonly_property, field_name)
                )
                setattr(self.__class__, field_name, prop)

        for i in range(0, jc["JRK_CONTROL_PIN_COUNT"]):
            for field_name, field_type in pin_info._fields_:
                prop = property(
                    fget=partial(self._get_pin_readonly_property, field_name, i)
                )
                setattr(self.pin_info[i].__class__, field_name, prop)

    @JED
    def _update_jrk_variables(self):
        e_p = self.jrklib.jrk_get_variables(
            byref(self._device_handle), byref(self._jrk_variables_p), c_bool(True)
        )
        self._jrk_variables = self._jrk_variables_p[0]
        return e_p

    def _get_jrk_readonly_property(self, field_name, _):
        self._update_jrk_variables()
        value = getattr(self._jrk_variables, field_name)
        if field_name == "error_flags_halting" or field_name == "error_flags_occurred":
            error_list = self._convert_error_bitmask(value)
            self._logger.debug(error_list)
        return value

    def _get_pin_readonly_property(self, field_name, pin_num, _):
        self._update_jrk_variables()
        return getattr(self._jrk_variables.pin_info[pin_num], field_name)

    def _convert_error_bitmask(self, e_bit_mask):
        ecodes = [
            "JRK_ERROR_AWAITING_COMMAND",
            "JRK_ERROR_NO_POWER",
            "JRK_ERROR_MOTOR_DRIVER",
            "JRK_ERROR_INPUT_INVALID",
            "JRK_ERROR_INPUT_DISCONNECT",
            "JRK_ERROR_FEEDBACK_DISCONNECT",
            "JRK_ERROR_SOFT_OVERCURRENT",
            "JRK_ERROR_SERIAL_SIGNAL",
            "JRK_ERROR_SERIAL_OVERRUN",
            "JRK_ERROR_SERIAL_BUFFER_FULL",
            "JRK_ERROR_SERIAL_CRC",
            "JRK_ERROR_SERIAL_PROTOCOL",
            "JRK_ERROR_SERIAL_TIMEOUT",
            "JRK_ERROR_HARD_OVERCURRENT",
        ]
        error_list = []
        for code in ecodes:
            if (e_bit_mask >> jc[code]) & 1:
                error_list.append(code)
        return error_list


class PyJrkEEPROMSettings(PyJrkSettingsBase):
    def __init__(self, device_handle, driver_handles, logger: LoggerProtocol):
        super().__init__(device_handle, driver_handles, logger)

    def _initialize_settings(self):
        """Get current settings from eeprom and fill the _local_settings"""
        self._get_eeprom_settings()
        self._local_settings = self._device_settings_p[0]

    def _convert_structure_to_properties(self):
        for field_name, field_type in jrk_settings._fields_:
            prop = property(
                fget=partial(self._get_jrk_setting_from_device, field_name),
                fset=partial(self._set_jrk_setting_with_option, field_name),
            )
            setattr(self.__class__, field_name, prop)

    def _get_jrk_setting_from_device(self, field_name: str, _):
        self._get_eeprom_settings()
        return getattr(self._device_settings, field_name)

    def apply(self):
        self._settings_fix()
        self._set_eeprom_settings()
        self._reinitialize()

    def print(self):
        settings_str = c_char_p()
        self._get_eeprom_settings()
        self._settings_to_string(settings_str)
        self._logger.debug(f"Device EEPROM settings:\n{settings_str.value.decode()}")


class PyJrkRAMSettings(PyJrkSettingsBase):
    def __init__(self, device_handle, driver_handles, logger: LoggerProtocol):
        super().__init__(device_handle, driver_handles, logger)
        self.auto_apply = True

    def _initialize_settings(self):
        """Get current settings from eeprom, fill the _local_settings with them
        and set the ram settings to the current eeprom settings"""
        self._get_eeprom_settings()
        self._local_settings = self._device_settings_p[0]
        self._set_ram_settings()

    def _get_jrk_setting_from_device(self, field_name: str, _):
        self._get_ram_settings()
        return getattr(self._device_settings, field_name)

    def apply(self):
        self._settings_fix()
        self._set_ram_settings()

    def print(self):
        settings_str = c_char_p()
        self._get_ram_settings()
        self._settings_to_string(settings_str)
        self._logger.debug(f"Device RAM settings:\n{settings_str.value.decode()}")


if __name__ == "__main__":
    jrk = PyJrk()
    print(jrk.list_connected_device_serial_numbers())



from ctypes import *

from pyjrk.pyjrk_protocol import jrk_constant as j_const


class libusbp_generic_interface(Structure):
    _fields_ = [
        ("interface_number", c_uint8),
        ("device_instance_id", c_char_p),
        ("filename", c_char_p),
    ]


class libusbp_generic_handle(Structure):
    # untested type, no HANDLE or WIN_INTERFACE_HANDLE type
    _fields_ = [("file_handle", c_ulong), ("winusb_handle", c_ulong)]


class libusbp_device(Structure):
    _fields_ = [
        ("device_instance_id", c_char_p),
        ("product_id", c_uint16),
        ("vendor_id", c_uint16),
        ("revision", c_uint16),
    ]


class jrk_device(Structure):
    _fields_ = [
        ("usb_device", POINTER(libusbp_device)),
        ("usb_interface", POINTER(libusbp_generic_interface)),
        ("serial_number", c_char_p),
        ("os_id", c_char_p),
        ("firmware_version", c_uint16),
        ("product", c_uint32),
    ]


class jrk_handle(Structure):
    _fields_ = [
        ("usb_handle", POINTER(libusbp_generic_handle)),
        ("device", POINTER(jrk_device)),
        ("cached_firmware_version_string", c_char_p),
    ]


class jrk_settings(Structure):
    _fields_ = [
        ("product", c_uint32),
        ("firmware_version", c_uint16),
        ("input_mode", c_uint8),
        ("input_error_minimum", c_uint16),
        ("input_error_maximum", c_uint16),
        ("input_minimum", c_uint16),
        ("input_maximum", c_uint16),
        ("input_neutral_minimum", c_uint16),
        ("input_neutral_maximum", c_uint16),
        ("output_minimum", c_uint16),
        ("output_neutral", c_uint16),
        ("output_maximum", c_uint16),
        ("input_invert", c_bool),
        ("input_scaling_degree", c_uint8),
        ("input_detect_disconnect", c_bool),
        ("input_analog_samples_exponent", c_uint8),
        ("feedback_mode", c_uint8),
        ("feedback_error_minimum", c_uint16),
        ("feedback_error_maximum", c_uint16),
        ("feedback_minimum", c_uint16),
        ("feedback_maximum", c_uint16),
        ("feedback_invert", c_bool),
        ("feedback_detect_disconnect", c_bool),
        ("feedback_dead_zone", c_uint8),
        ("feedback_analog_samples_exponent", c_uint8),
        ("feedback_wraparound", c_bool),
        ("serial_mode", c_uint8),
        ("serial_baud_rate", c_uint32),
        ("serial_timeout", c_uint32),
        ("serial_device_number", c_uint16),
        ("never_sleep", c_bool),
        ("serial_enable_crc", c_bool),
        ("serial_enable_14bit_device_number", c_bool),
        ("serial_disable_compact_protocol", c_bool),
        ("proportional_multiplier", c_uint16),
        ("proportional_exponent", c_uint8),
        ("integral_multiplier", c_uint16),
        ("integral_exponent", c_uint8),
        ("derivative_multiplier", c_uint16),
        ("derivative_exponent", c_uint8),
        ("pid_period", c_uint16),
        ("integral_divider_exponent", c_uint8),
        ("integral_limit", c_uint16),
        ("reset_integral", c_bool),
        ("pwm_frequency", c_uint8),
        ("current_samples_exponent", c_uint8),
        ("hard_overcurrent_threshold", c_uint8),
        ("current_offset_calibration", c_int16),
        ("current_scale_calibration", c_int16),
        ("motor_invert", c_bool),
        ("max_duty_cycle_while_feedback_out_of_range", c_uint16),
        ("max_acceleration_forward", c_uint16),
        ("max_acceleration_reverse", c_uint16),
        ("max_deceleration_forward", c_uint16),
        ("max_deceleration_reverse", c_uint16),
        ("max_duty_cycle_forward", c_uint16),
        ("max_duty_cycle_reverse", c_uint16),
        ("encoded_hard_current_limit_forward", c_uint16),
        ("encoded_hard_current_limit_reverse", c_uint16),
        ("brake_duration_forward", c_uint32),
        ("brake_duration_reverse", c_uint32),
        ("soft_current_limit_forward", c_uint16),
        ("soft_current_limit_reverse", c_uint16),
        ("soft_current_regulation_level_forward", c_uint16),
        ("soft_current_regulation_level_reverse", c_uint16),
        ("coast_when_off", c_bool),
        ("error_enable", c_uint16),
        ("error_latch", c_uint16),
        ("error_hard", c_uint16),
        ("vin_calibration", c_int16),
        ("disable_i2c_pullups", c_bool),
        ("analog_sda_pullup", c_bool),
        ("always_analog_sda", c_bool),
        ("always_analog_fba", c_bool),
        ("fbt_method", c_uint8),
        ("fbt_timing_clock", c_uint8),
        ("fbt_timing_polarity", c_bool),
        ("fbt_timing_timeout", c_uint16),
        ("fbt_samples", c_uint8),
        ("fbt_divider_exponent", c_uint8),
    ]


class pin_info(Structure):
    _fields_ = [
        ("analog_reading", c_uint16),
        ("digital_reading", c_bool),
        ("pin_state", c_uint8),
    ]


class jrk_variables(Structure):
    _fields_ = [
        ("input", c_uint16),
        ("target", c_uint16),
        ("feedback", c_uint16),
        ("scaled_feedback", c_uint16),
        ("integral", c_int16),
        ("duty_cycle_target", c_int16),
        ("duty_cycle", c_int16),
        ("current_low_res", c_uint8),
        ("pid_period_exceeded", c_bool),
        ("pid_period_count", c_uint16),
        ("error_flags_halting", c_uint16),
        ("error_flags_occurred", c_uint16),
        ("vin_voltage", c_uint16),
        ("current", c_uint16),
        ("device_reset", c_uint8),
        ("up_time", c_uint32),
        ("rc_pulse_width", c_uint16),
        ("fbt_reading", c_uint16),
        ("raw_current", c_uint16),
        ("encoded_hard_current_limit", c_uint16),
        ("last_duty_cycle", c_int16),
        ("current_chopping_consecutive_count", c_uint8),
        ("current_chopping_occurrence_count", c_uint8),
        ("force_mode", c_uint8),
        ("pin_info", pin_info * j_const["JRK_CONTROL_PIN_COUNT"]),
    ]


class jrk_error(Structure):
    _fields_ = [
        ("do_not_free", c_bool),
        ("message", c_char_p),
        ("code_count", c_size_t),
        ("code_array", POINTER(c_uint32)),
    ]
